==> slashdb_diag-1.0.0-py3-none-any.whl: slashdb_diag/config.py <==
from pathlib import Path
from typing import Optional, Union, Set, Dict, List
from io import StringIO
import configparser
import re

import yaml

from .files import DiagFile
from .secrets import create_secret_replacement, ApiKeyConfig

FILE_HANDLERS = [
    "handlers.FileHandler",
    "handlers.WatchedFileHandler",
    "handlers.RotatingFileHandler",
    "handlers.TimedRotatingFileHandler",
]


def read_slashdb_config(path: Path) -> configparser.ConfigParser:
    here = path.parent.absolute()
    config = configparser.ConfigParser(defaults={"here": here})
    config.read(path)
    return config


def get_log_paths_from_config(path: Union[Path, str]) -> Set[Path]:
    config = read_slashdb_config(Path(path))

    handler_names = config["handlers"]["keys"].split(", ")
    handler_section_names = [f"handler_{n}" for n in handler_names]

    log_paths = {get_uwsgi_logfile_path(config["uwsgi"])}

    parsed_handler_paths = (
        parse_handler_section(config[name]) for name in handler_section_names
    )
    filtered_handler_paths: Set[Path] = set(filter(None, parsed_handler_paths))

    log_paths |= filtered_handler_paths

    return log_paths


def get_uwsgi_logfile_path(section) -> Path:
    if path := section.get("logto"):
        return Path(path)
    if match := re.match("file:(?:.*=)?(\/[\w/.]*)", section["logger"]):
        return Path(match.groups()[0])
    raise Exception("Failed to get log path for uWSGI")


def parse_handler_section(section) -> Optional[Path]:
    if section["class"] not in FILE_HANDLERS:
        return None

    # args = '("/var/log/slashdb/slashdb.log",\'a\', \'utf-8\')'
    args = section["args"]
    # first argument is the filepath
    path = args[1:].split(",")[0][1:-1]
    return Path(path)


def get_nginx_log_paths_from_config(path: Union[str, Path]) -> Set[Path]:
    with open(path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    paths = set()
    for line in lines:
        line = line.strip()
        r = re.match("(?:error_log|access_log) (\/[\w\/.]*)[ ;]", line)
        if r:
            path = Path(r.groups()[0])
            if path.is_absolute():
                paths.add(path)
    return paths


def get_nginx_log_format_from_config(path: Union[str, Path]) -> Optional[str]:
    with open(path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    for i, line in enumerate(lines):
        if line.strip().startswith("log_format"):
            # because the format string can be spread on multiple lines,
            # we have to do such wizardry to find it and convert
            # to a single string
            format_lines = line.strip()
            for line in lines[i + 1 :]:
                format_lines += line.strip()
                if line[-2] == ";":
                    break

            # find where the first string starts
            str_idx_a = format_lines.find('"')
            if str_idx_a == -1:
                str_idx_a = len(format_lines)
            str_idx_b = format_lines.find("'")
            if str_idx_b == -1:
                str_idx_b = len(format_lines)
            str_idx = min(str_idx_a, str_idx_b)

            # get all the strings, remove '' and "" and first and last " or '
            return format_lines[str_idx + 1 : -2].replace("''", "").replace('""', "")

    return None


def get_apikeys_from_config(path: Union[str, Path]) -> ApiKeyConfig:
    config = read_slashdb_config(Path(path))
    app_config = config["app:sdbapp"]
    apikeys = set()
    if users_path := app_config.get("users_path"):
        apikeys = set(filter(None, get_apikeys_from_user_config(users_path)))

    apikey_name = app_config.get("apikey_name", "apikey")

    return ApiKeyConfig(apikeys, apikey_name)


def get_configs_from_config(path: Union[str, Path]) -> List[DiagFile]:
    configs = []
    config = read_slashdb_config(Path(path))

    redacted_config = redact_slashdb_ini(config)
    ini_str = StringIO()
    redacted_config.write(ini_str)
    ini_lines = ini_str.getvalue().splitlines(keepends=True)
    configs.append(DiagFile("slashdb.ini", ini_lines))

    app_config = config["app:sdbapp"]

    users_path = app_config["users_path"]
    if users_config := read_yaml(users_path):
        users_config = redact_users_config(users_config)
        users_config_lines = yaml.dump(users_config).splitlines(keepends=True)
        configs.append(DiagFile(Path(users_path).name, users_config_lines))

    databases_path = app_config["databases_path"]
    if databases_config := read_yaml(databases_path):
        databases_config = redact_databases_config(databases_config)
        databases_config_lines = yaml.dump(databases_config).splitlines(keepends=True)
        configs.append(DiagFile(Path(databases_path).name, databases_config_lines))

    configs_rest = ["querydefs_path", "license", "auth_config_path"]
    configs.extend((DiagFile.open(Path(app_config[key])) for key in configs_rest))

    configs = list(filter(None, configs))

    return configs


def read_yaml(path: str) -> Dict:
    try:
        with open(path, "r", encoding="utf-8") as file:
            return yaml.load(file, Loader=yaml.Loader)
    except Exception as e:
        print(f"Failed to open file {str(path)}: {e}")
        return None


def get_apikeys_from_user_config(path: Union[Path, str]) -> Set[str]:
    apikeys = set()
    try:
        with open(path, "r", encoding="utf-8") as file:
            config: dict = yaml.load(file, Loader=yaml.Loader)
        for user in config.values():
            if api_key := user.get("api_key"):
                apikeys.add(api_key)
    except Exception as e:
        print(f"Failed to open file {str(path)}: {e}")
    return apikeys


def get_config_paths_from_config(path: Path) -> Set[Path]:
    KEYS = [
        "databases_path",
        "users_path",
        "querydefs_path",
        "license",
        "auth_config_path",
    ]
    config = read_slashdb_config(path)
    app_config = config["app:sdbapp"]
    return {Path(app_config[key]) for key in KEYS}


REDACT_STR = "REDACTED"


def redact_slashdb_ini(config: configparser.ConfigParser) -> configparser.ConfigParser:
    try:
        if "session.secret" in config["app:sdbapp"]:
            config["app:sdbapp"]["session.secret"] = REDACT_STR
    except KeyError:
        pass
    try:
        if "secret" in config["app:sdbapp"]:
            config["app:sdbapp"]["secret"] = REDACT_STR
    except KeyError:
        pass
    return config


def redact_users_config(config: Dict) -> Dict:
    for user in config.values():
        if api_key := user.get("api_key"):
            user["api_key"] = create_secret_replacement(api_key)
        try:
            if not user["databases"]:
                continue
            for db in user["databases"].values():
                if not db:
                    continue
                if "dbpass" in db.keys():
                    db["dbpass"] = REDACT_STR
        except KeyError:
            pass
    return config


def redact_databases_config(config: Dict) -> Dict:
    for db in config.values():
        try:
            _ = db["autoload_user"]["dbpass"]
            db["autoload_user"]["dbpass"] = REDACT_STR
        except KeyError:
            pass
    return config

==> slashdb_diag-1.1.0-py3-none-any.whl: slashdb_diag/config.py <==
from pathlib import Path
from typing import Optional, Union, Set, Dict, List
from io import StringIO
import configparser
import re
from itertools import zip_longest

import yaml
from slashdb.cli.utils import read_slashdb_config, read_yaml

from .files import DiagFile
from .secrets import create_secret_replacement, ApiKeyConfig

FILE_HANDLERS = [
    "handlers.FileHandler",
    "handlers.WatchedFileHandler",
    "handlers.RotatingFileHandler",
    "handlers.TimedRotatingFileHandler",
]

DEFAULT_LOG_HEADER = (
    "date & time | timestamp | source | logger name | level | pid | thread"
    " name | request id | request method | URL | resp status code | "
    "request processing time | request size (header + body) | response "
    "size (headers + body) | remote address | http referer | http user "
    "agent | python module | user id | message"
)

LOG_COLUMN_MAPPING = {
    "${strftime:%%Y-%%m-%%d %%H:%%M:%%S}": "date & time",
    "%(asctime)-.19s": "date & time",
    "${millis}": "timestamp",
    "%(created)d%(msecs)03d": "timestamp",
    "sdbapp": "source",
    "%(name)s": "logger name",
    "%(levelname)-5.5s": "level",
    "%(process)d": "pid",
    "%(threadName)s": "thread name",
    "%(request_id)s": "request id",
    "%(request_method)s": "request method",
    "%(request_url)s": "URL",
    "%(response_status_code)s": "resp status code",
    "%(secs)": "request processing time",
    "$request_length": "request size (header + body)",
    "%(size)": "response size (headers + body)",
    "%(request_client_addr)s": "remote address",
    "%(request_referer)s": "http referer",
    "%(request_user_agent)s": "http user agent",
    "%(module)s:%(funcName)s:%(lineno)d": "python module",
    "%(user_id)s": "user id",
    "%(message)s": "message",
}

MISSING_LOG_COLUMN_NAME = "UNKNOWN"


def get_log_paths_from_config(path: Union[Path, str]) -> Set[Path]:
    config = read_slashdb_config(Path(path))

    handler_names = config["handlers"]["keys"].split(", ")
    handler_section_names = [f"handler_{n}" for n in handler_names]

    log_paths = {get_uwsgi_logfile_path(config["uwsgi"])}

    parsed_handler_paths = (
        parse_handler_section(config[name]) for name in handler_section_names
    )
    filtered_handler_paths: Set[Path] = set(filter(None, parsed_handler_paths))

    log_paths |= filtered_handler_paths

    return log_paths


def get_uwsgi_logfile_path(section) -> Path:
    if path := section.get("logto"):
        return Path(path)
    if match := re.match(r"file:(?:.*=)?(\/[\w/.]*)", section["logger"]):
        return Path(match.groups()[0])
    raise Exception("Failed to get log path for uWSGI")


def parse_handler_section(section) -> Optional[Path]:
    if section["class"] not in FILE_HANDLERS:
        return None

    # args = '("/var/log/slashdb/slashdb.log",\'a\', \'utf-8\')'
    args = section["args"]
    # first argument is the filepath
    path = args[1:].split(",")[0][1:-1]
    return Path(path)


def get_nginx_log_paths_from_config(path: Union[str, Path]) -> Set[Path]:
    with open(path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    paths = set()
    for line in lines:
        line = line.strip()
        r = re.match(r"(?:error_log|access_log) (\/[\w\/.]*)[ ;]", line)
        if r:
            path = Path(r.groups()[0])
            if path.is_absolute():
                paths.add(path)
    return paths


def get_nginx_log_format_from_config(path: Union[str, Path]) -> Optional[str]:
    with open(path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    for i, line in enumerate(lines):
        if line.strip().startswith("log_format"):
            # because the format string can be spread on multiple lines,
            # we have to do such wizardry to find it and convert
            # to a single string
            format_lines = line.strip()
            for line in lines[i + 1 :]:
                format_lines += line.strip()
                if line[-2] == ";":
                    break

            # find where the first string starts
            str_idx_a = format_lines.find('"')
            if str_idx_a == -1:
                str_idx_a = len(format_lines)
            str_idx_b = format_lines.find("'")
            if str_idx_b == -1:
                str_idx_b = len(format_lines)
            str_idx = min(str_idx_a, str_idx_b)

            # get all the strings, remove '' and "" and first and last " or '
            return format_lines[str_idx + 1 : -2].replace("''", "").replace('""', "")

    return None


def get_log_formats_from_ini(path: Union[str, Path]) -> Set[str]:
    config = read_slashdb_config(Path(path), interpolation=None)

    formats = set()

    format_prefix = "format "

    if config.has_option("uwsgi", "log-encoder"):
        uwsgi_internal_format = config["uwsgi"]["log-encoder"]
        if uwsgi_internal_format.startswith(format_prefix):
            uwsgi_internal_format = uwsgi_internal_format[len(format_prefix) :]
            formats.add(uwsgi_internal_format)

    if config.has_option("uwsgi", "log-req-encoder"):
        uwsgi_request_format = config["uwsgi"]["log-req-encoder"]
        if config.has_option("uwsgi", "log-format"):
            uwsgi_request_format = uwsgi_request_format.replace(
                "${msgnl}", config["uwsgi"]["log-format"]
            )
        if uwsgi_request_format.startswith(format_prefix):
            uwsgi_request_format = uwsgi_request_format[len(format_prefix) :]
            formats.add(uwsgi_request_format)

    if config.has_option("formatters", "keys"):
        python_formatters_names = config["formatters"]["keys"].split(", ")

        def _get_format(formatter_name: str) -> str:
            return config[f"formatter_{formatter_name}"]["format"]

        formats.update(map(_get_format, python_formatters_names))

    return formats


def generate_log_header(formats: Set[str]) -> str:
    if len(formats) == 0:
        return DEFAULT_LOG_HEADER

    def _split_columns(format: str) -> List[str]:
        return list(map(str.strip, format.split("|")))

    folded = list(map(list, zip_longest(*map(_split_columns, formats), fillvalue="")))

    def _map_column_name(column_names: List[str]) -> str:
        for name in column_names:
            if mapped := LOG_COLUMN_MAPPING.get(name):
                return mapped
        return MISSING_LOG_COLUMN_NAME

    mapped = map(_map_column_name, folded)

    formatted = " | ".join(mapped)

    return formatted


def get_apikeys_from_config(path: Union[str, Path]) -> ApiKeyConfig:
    config = read_slashdb_config(Path(path))
    app_config = config["app:sdbapp"]
    apikeys = set()
    if users_path := app_config.get("users_path"):
        apikeys = set(filter(None, get_apikeys_from_user_config(users_path)))

    apikey_name = app_config.get("apikey_name", "apikey")

    return ApiKeyConfig(apikeys, apikey_name)


def get_configs_from_config(path: Union[str, Path]) -> List[DiagFile]:
    configs = []
    config = read_slashdb_config(Path(path))

    redacted_config = redact_slashdb_ini(config)
    ini_str = StringIO()
    redacted_config.write(ini_str)
    ini_lines = ini_str.getvalue().splitlines(keepends=True)
    configs.append(DiagFile("slashdb.ini", ini_lines))

    app_config = config["app:sdbapp"]

    users_path = app_config["users_path"]
    if users_config := read_yaml(users_path):
        users_config = redact_users_config(users_config)
        users_config_lines = yaml.dump(users_config).splitlines(keepends=True)
        configs.append(DiagFile(Path(users_path).name, users_config_lines))

    databases_path = app_config["databases_path"]
    if databases_config := read_yaml(databases_path):
        databases_config = redact_databases_config(databases_config)
        databases_config_lines = yaml.dump(databases_config).splitlines(keepends=True)
        configs.append(DiagFile(Path(databases_path).name, databases_config_lines))

    configs_rest = ["querydefs_path", "license", "auth_config_path"]
    configs.extend((DiagFile.open(Path(app_config[key])) for key in configs_rest))

    configs = list(filter(None, configs))

    return configs


def get_apikeys_from_user_config(path: Union[Path, str]) -> Set[str]:
    apikeys = set()
    try:
        with open(path, "r", encoding="utf-8") as file:
            config: dict = yaml.load(file, Loader=yaml.Loader)
        for user in config.values():
            if api_key := user.get("api_key"):
                apikeys.add(api_key)
    except Exception as e:
        print(f"Failed to open file {str(path)}: {e}")
    return apikeys


def get_config_paths_from_config(path: Path) -> Set[Path]:
    KEYS = [
        "databases_path",
        "users_path",
        "querydefs_path",
        "license",
        "auth_config_path",
    ]
    config = read_slashdb_config(path)
    app_config = config["app:sdbapp"]
    return {Path(app_config[key]) for key in KEYS}


REDACT_STR = "REDACTED"


def redact_slashdb_ini(config: configparser.ConfigParser) -> configparser.ConfigParser:
    try:
        if "session.secret" in config["app:sdbapp"]:
            config["app:sdbapp"]["session.secret"] = REDACT_STR
    except KeyError:
        pass
    try:
        if "secret" in config["app:sdbapp"]:
            config["app:sdbapp"]["secret"] = REDACT_STR
    except KeyError:
        pass
    return config


def redact_users_config(config: Dict) -> Dict:
    for user in config.values():
        if api_key := user.get("api_key"):
            user["api_key"] = create_secret_replacement(api_key)
        try:
            if not user["databases"]:
                continue
            for db in user["databases"].values():
                if not db:
                    continue
                if "dbpass" in db.keys():
                    db["dbpass"] = REDACT_STR
        except KeyError:
            pass
    return config


def redact_databases_config(config: Dict) -> Dict:
    for db in config.values():
        try:
            _ = db["autoload_user"]["dbpass"]
            db["autoload_user"]["dbpass"] = REDACT_STR
        except KeyError:
            pass
    return config
